/*--------------------------------------------------------------------------*/
/*
 * This program contains USB Skeleton driver's code originally written 
 * by Greg Kroah-Hartman, Copyright (C) 2001.  Thank you Greg.
 *
 * Copyright (C) 2003 by Kazuo Nakayoshi <kazuo.nakayoshi@kek.jp>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*--------------------------------------------------------------------------*/
/*
 * USB Device Driver for TOYO CC/7700 USB Adapter
 *
 * HISTORY
 *
 *   3 May, 2003 First release(K)
 *
 * $Id: ccusb.c,v 1.1.1.1 2003/08/25 02:17:30 nakayosi Exp $
 */
/*--------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/wrapper.h>
#include <linux/usb.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include "ccusb.h"

#define USB_CCUSB_MINOR_BASE    200
#define USB_CCUSB_VENDOR_ID  0x0BE3  /* TOYO Corp. */
#define USB_CCUSB_PRODUCT_ID 0x0020  /* CC/USB */
#define MAX_DEVICES    8

static int debug = 0;
#define dbg(format, arg...) do { if (debug) printk(KERN_DEBUG __FILE__ ": " format "\n" , ## arg); } while (0)

static void ccusb_write_bulk_callback (struct urb *urb);
/*--------------------------------------------------------------------------*/
struct ccusb_data{
	struct usb_device       *udev;
	struct usb_interface    *interface;
	unsigned char           minor;
	unsigned char           adapter_id;
	struct urb *		bulk_write_urb;
	int                     open_count;
	wait_queue_head_t       wait_queue;
	int                     single_in_ep_addr;
	int                     single_out_ep_addr;
	int                     block_in_ep_addr;
	int                     block_out_ep_addr;
	unsigned char           *bulk_in_buffer;
	unsigned char           *bulk_out_buffer;
	struct semaphore        sem;
};

static struct ccusb_data *minor_table[MAX_DEVICES];

static DECLARE_MUTEX (minor_table_mutex);
/*--------------------------------------------------------------------------*/
#define DEBUG_CCUSB
/*--------------------------------------------------------------------------*/
static inline void ccusb_delete (struct ccusb_data *ccusb)
{
	dbg(__FUNCTION__ " ccusb_delete: ccusb->adapter_id#%d\n", 
	    ccusb->adapter_id);
	minor_table[ccusb->adapter_id] = NULL;
	if (ccusb->bulk_in_buffer != NULL)
		kfree (ccusb->bulk_in_buffer);

	if (ccusb->bulk_out_buffer != NULL)
		kfree (ccusb->bulk_out_buffer);

	if (ccusb->bulk_write_urb != NULL)
		usb_free_urb (ccusb->bulk_write_urb);
	kfree (ccusb);
}
/*--------------------------------------------------------------------------*/
static unsigned int
usb_read(struct ccusb_data *ccusb, void * buf, size_t len, int ep_addr)
{
	size_t retlen = 0;
	int retval = 0;
	
	dbg("%s\n", __FUNCTION__);

	if (ccusb->udev == NULL){
		printk("ERROR: usb_read: ccusb->udev == NULL");
		retval = -ENODEV;
		goto exit;
	}

	retval = usb_bulk_msg ( ccusb->udev, 
	                        usb_rcvbulkpipe ( ccusb->udev, ep_addr),
				buf, 
				len,
				&retlen, HZ*4);
	dbg(__FUNCTION__ " usb_bulk_msg: retval = 0x%x(%d)\n", retval, retval);
	dbg(__FUNCTION__ " usb_read: retlen = 0x%x\n", retlen);
	if(!retval) {
		retval = retlen;
	}
 exit:
	return retval;
}

/*--------------------------------------------------------------------------*/
static unsigned int
usb_write(struct ccusb_data *ccusb, unsigned char *buf, size_t count, int ep_addr)
{
	ssize_t bytes_written = 0;
	int retval = 0;
	
	dbg("%s\n", __FUNCTION__);

	if (ccusb->udev == NULL){
		err(__FUNCTION__ "ccusb->udev == NULL");
		retval = -ENODEV;
		goto exit;
	}

	bytes_written = (count > MAX_TRANS_BYTES) ? MAX_TRANS_BYTES : count;
	dbg(__FUNCTION__ "bytes_written = 0x%x\n",bytes_written);

	FILL_BULK_URB (ccusb->bulk_write_urb, ccusb->udev,
		       usb_sndbulkpipe(ccusb->udev, ep_addr),
		       buf,
		       bytes_written,
		       ccusb_write_bulk_callback, ccusb);
	retval = usb_submit_urb(ccusb->bulk_write_urb);
	
	if (retval) {
		err(__FUNCTION__ " failed submitting write urb %d\n", retval);
	} else {
		retval = bytes_written;
	}
 exit:
	return retval;
}
/*--------------------------------------------------------------------------*/

static unsigned int
usb_write_wait(struct ccusb_data *ccusb, unsigned char *buffer, 
	       size_t count, int ep_addr)
{
	ssize_t bytes_written = 0;
	int retval = 0;
	
	dbg(__FUNCTION__);
	if (ccusb->udev == NULL){
		err(__FUNCTION__ " ccusb->udev == NULL");
		retval = -ENODEV;
		goto exit;
	}

	bytes_written = (count > MAX_TRANS_BYTES) ? MAX_TRANS_BYTES : count;
	dbg(__FUNCTION__ " bytes_written = 0x%x\n", bytes_written);

	retval = usb_bulk_msg ( ccusb->udev, 
	                        usb_sndbulkpipe ( ccusb->udev, ep_addr),
				buffer, 
				count,
				&bytes_written, HZ*4);

	if (retval) {
		err(__FUNCTION__ " failed submitting write urb %d\n", retval);
	} else {
		retval = bytes_written;
	}
 exit:
	return retval;
}
/*--------------------------------------------------------------------------*/

static int
do_camac_single( struct ccusb_data *ccusb, struct camac_single * cam_sing, 
		 size_t word_size )
{
	unsigned char cam_com[0x7];
	unsigned char cam_dat[0x4];
	unsigned char cam_stat;
	size_t com_size = 0; 
	int func, ret, retval = 0; 

	dbg(__FUNCTION__);

	if (ccusb->udev == NULL){
		retval = -ENODEV;
		goto exit;
	}

	cam_com[1] = cam_sing->naf[0] & 0x1f;
	cam_com[2] = cam_sing->naf[1] & 0x0f;
	cam_com[3] = cam_sing->naf[2] & 0x1f;
	
	func = cam_com[3];
	dbg(__FUNCTION__ " N = 0x%x\n", cam_com[1]);
	dbg(__FUNCTION__ " A = 0x%x\n", cam_com[2]);
	dbg(__FUNCTION__ " F = 0x%x\n", cam_com[3]);
	dbg(__FUNCTION__ " word size = %d\n", word_size);

	if( func < 8 ) {
		com_size  = 4;
		if( word_size == 2 ) {
			cam_com[0] = SINGLE_STAT;
			cam_com[4] = 0x0;
			cam_com[5] = 0x0;
		}
		else if( word_size == 3 ) {
			cam_com[0] = SINGLE_STAT | CAM_LBIT;
			cam_com[4] = 0x0;
			cam_com[5] = 0x0;
			cam_com[6] = 0x0;
		}

		ret = usb_write(ccusb, cam_com, com_size, 
				ccusb->single_out_ep_addr);
		dbg(__FUNCTION__ " usb_write ret = 0x%x\n", ret);
		if(!ret){
			err(__FUNCTION__ " write to USB failed\n");
			retval = -EFAULT;
			goto exit;
		}

		ret = usb_read(ccusb, cam_dat, (word_size + 1), //data + stat
				       ccusb->single_in_ep_addr);
		dbg(__FUNCTION__ " usb_read ret = 0x%x\n", ret);
		if(ret <= 0){
			err(__FUNCTION__ " read from USB failed: %d(0x%x)\n",
			       ret,ret);
			retval = -EFAULT;
			goto exit;
		}
		
		if(word_size == 2) {
			if(cam_dat[2] & ERRORS) {
				retval = -EFAULT;
				err(__FUNCTION__ " camac status = 0x%x\n", 
				    cam_dat[2]);
				goto exit;
			}else {
				cam_sing->data = * (unsigned short *)cam_dat;
				cam_sing->qx   = (~cam_dat[2] & 0x03);
			}
		}else if(word_size == 3) {
			if(cam_dat[3] & ERRORS) {
				retval = -EFAULT;
				err(__FUNCTION__ " camac status = 0x%x\n", 
				    cam_dat[3]);
				goto exit;
			}else {
				cam_sing->data = * (unsigned int *)cam_dat;
				cam_sing->qx   = (~cam_dat[3] & 0x03);
			}
		}
		dbg(__FUNCTION__ " cam_sing-data = %x\n",cam_sing->data);
		dbg(__FUNCTION__ " cam_qx = %d\n", (~cam_dat[2] & 0x03));

	}

	else if( func > 15 && func < 24 ){
		dbg(__FUNCTION__ " cam_sing w dat = %x\n",cam_sing->data);
		if( word_size == 2 ) {
			cam_com[0] = SINGLE_STAT;
			cam_com[4] = cam_sing->data & 0x00ff;
			cam_com[5] = (cam_sing->data & 0xff00) >> 8;
			com_size  = 6;
		}
		else if( word_size == 3 ) {
			cam_com[0] = SINGLE_STAT | CAM_LBIT;
			cam_com[4] =  cam_sing->data & 0x0000ff;
			cam_com[5] = (cam_sing->data & 0x00ff00) >>  8;
			cam_com[6] = (cam_sing->data & 0xff0000) >> 16;
			com_size  = 7;
		}
		ret = usb_write(ccusb, cam_com, com_size, 
				ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " write to USB failed\n");
			retval = -EFAULT;
				goto exit;
		}
			
		ret = usb_read(ccusb, &cam_stat, sizeof(cam_stat), 
			       ccusb->single_in_ep_addr);
		if(ret <= 0){
			err(__FUNCTION__ " read from USB failed\n");
			retval = -EFAULT;
			goto exit;
		}
		if(cam_stat & ERRORS) {
			retval = -EFAULT;
			err(__FUNCTION__ " camac status = 0x%x\n", cam_stat);
			goto exit;
		}else {
			cam_sing->qx   = (~cam_stat & 0x03);
		}
		dbg(__FUNCTION__ " cam_stat = %d\n",cam_stat);
		dbg(__FUNCTION__ " cam_stat = %d\n", (~cam_stat & 0x03));
		
	}
	else {
		dbg(__FUNCTION__ " cam_sing no dat  f = %x\n", func);
		com_size  = 4;
		if( word_size == 2 ) {
			cam_com[0] = SINGLE_STAT;
		}
		else if( word_size == 3 ) {
			cam_com[0] = SINGLE_STAT | CAM_LBIT;
		}
		ret = usb_write(ccusb, cam_com, com_size, 
				ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " write to USB failed\n");
			retval = -EFAULT;
				goto exit;
		}
			
		ret = usb_read(ccusb, &cam_stat, sizeof(cam_stat), 
			       ccusb->single_in_ep_addr);
		if(ret <= 0){
			err(__FUNCTION__ " read from USB failed:  %d(0x%x)\n", 
			    ret, ret);
			retval = -EFAULT;
			goto exit;
		}
		if(cam_stat & ERRORS) {
			retval = -EFAULT;
			err(__FUNCTION__ " camac status = 0x%x\n", cam_stat);
			goto exit;
		}else {
			cam_sing->qx   = (~cam_stat & 0x03);
		}	
		dbg(__FUNCTION__ " cam_stat = %d\n",cam_stat);
		dbg(__FUNCTION__ " ~cam_stat & 0x03 = %d\n", 
		    (~cam_stat & 0x03));
	}
	
 exit:

	return retval;
}

/*--------------------------------------------------------------------------*/

static int
do_camac_block( struct ccusb_data *ccusb, struct camac_block * cam_blk, 
		size_t word_size )
{
	unsigned char cam_com[0x06];
	unsigned char blk_stat[0x03];
	unsigned char stat_com = READ_STAT;
	unsigned int word_cnt;
	unsigned int mode;
	unsigned int func;
	int ret, retval = 0; 

	dbg(__FUNCTION__);

	if (ccusb->udev == NULL){
		retval = -ENODEV;
		goto exit;
	}

	if(cam_blk->ch > MAX_TRANS_BYTES / word_size) {
		err(__FUNCTION__ " Maximum Transfer WORDS exceeded\n");
		retval = -EFAULT;
		goto exit;
	}
	else
		word_cnt = cam_blk->ch;

	dbg(__FUNCTION__ "word_cnt = 0x%x\n", word_cnt);

	mode = cam_blk->mode;
	func = cam_blk->naf[0x02] & 0x1f;
	dbg(__FUNCTION__ "func = 0x%x\n", func);
	
	if(word_size == 2)
		cam_com[0] = mode;
	else if(word_size == 3)
		cam_com[0] = mode | CAM_LBIT;	

	if( mode == BLOCK_ASCAN ) { //Address Scan Mode
		cam_com[1] = cam_blk->naf[0] & 0x1f; //Start Address
		cam_com[2] = cam_blk->naf[1] & 0x1f; //Stop Address
		cam_com[3] = func;                   //CAMAC f
	} else { //Not Address Scan Mode
		cam_com[1] = cam_blk->naf[0] & 0x1f; //CAMAC N
		cam_com[2] = cam_blk->naf[1] & 0x0f; //CAMAC A
		cam_com[3] = func;                   //CAMAC F
	}
	
	cam_com[4] = word_cnt & 0x00ff;
	cam_com[5] = ( word_cnt & 0xff00 ) >> 8;
	
	if( func >= 0 && func < 8 ){
		if( word_cnt == 0){
			err(__FUNCTION__ " Trans. Word Count was ZERO\n");
			retval = -EINVAL;
			goto exit;
		}
                          
		ret = usb_write(ccusb, cam_com, sizeof(cam_com), 
				ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " USB Write failed\n");
			retval = -EFAULT;
			goto exit;
		}
		
		ret = usb_read(ccusb, ccusb->bulk_in_buffer, 
			       word_cnt*word_size, 
			       ccusb->block_in_ep_addr);

		if(!ret){
			err(__FUNCTION__ " USB Read failed\n");
			retval = -EFAULT;
			goto exit;
		}

		ret = usb_write(ccusb, &stat_com, sizeof(stat_com), 
				ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " USB Write failed\n");
			retval = -EFAULT;
			goto exit;
		}

		ret = usb_read(ccusb, blk_stat, sizeof(blk_stat), 
			       ccusb->single_in_ep_addr);
		if(!ret){
			err(__FUNCTION__ " USB Read failed\n");
			retval = -EFAULT;
			goto exit;
		}

		if( blk_stat[0] & ERRORS ) {
			err(__FUNCTION__ " BLK Read status = 0x%x\n", 
			    blk_stat[0]);
			retval = -EFAULT;
			goto exit;
		}
		
		retval = (blk_stat[1] + ( blk_stat[2] << 8 ) ) * word_size;
		
	}
	else if( func > 15 && func < 24 ){
                        
		if( word_cnt == 0){
			retval = -EINVAL;
			err(__FUNCTION__ " Trans. Word Count was ZERO\n");
			goto exit;
		}

		if( copy_from_user(ccusb->bulk_out_buffer, 
				   (unsigned short *)cam_blk->data, 
				   word_cnt*word_size) ){
                          retval = -EFAULT;
                          err(__FUNCTION__ "copy_from_user failed\n");
                          goto exit;
                        }
                          
		ret = usb_write(ccusb, cam_com, sizeof(cam_com), 
				ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " USB Write failed ret = %d\n", ret);
			retval = -EFAULT;
			goto exit;
		}
		
		ret = usb_write_wait(ccusb, ccusb->bulk_out_buffer, 
		//ret = usb_write(ccusb, ccusb->bulk_out_buffer, 
			       word_cnt*word_size, 
			       ccusb->block_out_ep_addr);

		if(!ret){
			err(__FUNCTION__ " USB Write failed ret = %d\n\n", ret);
			retval = -EFAULT;
			goto exit;
		}

		ret = usb_write(ccusb, &stat_com, sizeof(stat_com), 
				ccusb->single_out_ep_addr);
		if(!ret){
		        err(__FUNCTION__ " USB Write failed\n");
			retval = -EFAULT;
			goto exit;
		}

		ret = usb_read(ccusb, blk_stat, sizeof(blk_stat), 
			       ccusb->single_in_ep_addr);
		if(!ret){
			err(__FUNCTION__ " USB Read failed\n");
			retval = -EFAULT;
			goto exit;
		}

		if( blk_stat[0] & ERRORS ) {
			err(__FUNCTION__ " BLK Write status = 0x%x\n", 
			    blk_stat[0]);
			retval = -EFAULT;
			goto exit;
		}
		
		retval = (blk_stat[1] + ( blk_stat[2] << 8 ) ) * word_size;
	}
 exit:
	return retval;
}

static int read_id(struct ccusb_data * ccusb)
{
	int ret, retval;
	unsigned char camac_com = READ_ID;
	unsigned char adapter_id;

	dbg(__FUNCTION__);

	if (ccusb->udev == NULL) {
		err(__FUNCTION__ " ccusb->udev not exist\n");
		return -ENODEV;
	}

	ret = usb_write(ccusb, &camac_com, sizeof(camac_com), 
			ccusb->single_out_ep_addr);
	if(!ret){
		err(__FUNCTION__ " write to USB failed\n");
		retval = -EFAULT;
		goto exit;
	}
	ret = usb_read(ccusb, &adapter_id, sizeof(adapter_id), 
                               ccusb->single_in_ep_addr);
	if(ret <= 0){
		err(__FUNCTION__ " read from USB failed\n");
		retval = -EFAULT;
		goto exit;
	}

	dbg(__FUNCTION__ " adapter_id = 0x%x\n", adapter_id);
	retval = adapter_id;

 exit:
	return retval;
}

static int read_status(struct ccusb_data * ccusb)
{
	int ret, retval;
	unsigned char camac_com = READ_STAT;
	unsigned char camac_stat;

	dbg(__FUNCTION__);

	if (ccusb->udev == NULL) {
		printk("ERROR: read_status: ccusb->udev not exist\n");
		return -ENODEV;
	}

	ret = usb_write(ccusb, &camac_com, sizeof(camac_com), 
			ccusb->single_out_ep_addr);
	if(!ret){
		err(__FUNCTION__ " write to USB failed\n");
		retval = -EFAULT;
		goto exit;
	}
	ret = usb_read(ccusb, &camac_stat, sizeof(camac_stat), 
                               ccusb->single_in_ep_addr);
	if(ret <= 0){
		err(__FUNCTION__ " read from USB failed\n");
		retval = -EFAULT;
		goto exit;
	}

	dbg(__FUNCTION__ " camac status = 0x%x\n", camac_stat);
	retval = camac_stat;

 exit:
	return retval;
}

/*--------------------------------------------------------------------------*/
static int
ccusb_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	    unsigned long arg)
{
	struct ccusb_data    *ccusb = NULL;
	struct camac_single *cam_sing = NULL;
	struct camac_block  *cam_blk = NULL;
	int retval = 0;

	dbg(__FUNCTION__);
	ccusb = (struct ccusb_data *)file->private_data;

	down (&ccusb->sem);

	if (ccusb->udev == NULL) {
		up (&ccusb->sem);
		err(__FUNCTION__ " ccusb->udev not exist");
		return -ENODEV;
	}
 
	switch (cmd) {

	case CAM_STATUS:
	{
		dbg(__FUNCTION__ "CAM_STAT");
		retval = read_status( ccusb );

		break;
	}
	case CAM_WORD:
	{
		size_t word_siz = 2;

		dbg(__FUNCTION__ "CAM_WORD");
		cam_sing = kmalloc(sizeof(struct camac_single), GFP_KERNEL);
		if(cam_sing == NULL) {
			err(__FUNCTION__ " Out of Memory");
			retval = -ENOMEM;
		}
		if( copy_from_user(cam_sing, (void *) arg, 
				   sizeof(struct camac_single)) ){
			retval = -EFAULT;
			err(__FUNCTION__ " copy_from_user failed");
			goto exit;
		}

		retval = do_camac_single( ccusb, cam_sing, word_siz );
		if(retval < 0) {
			err(__FUNCTION__ " do_camac_single failed");
			goto exit;
		}
		if( copy_to_user((void*)arg ,cam_sing, 
				 sizeof(struct camac_single)) ) {
			retval = -EFAULT;
			err(__FUNCTION__ " copy_to_user failed");
			goto exit;
		}
		break;
	}
	case CAM_LONG:
	{
		size_t word_siz = 3;

		dbg(__FUNCTION__ "CAM_LONG");
		cam_sing = kmalloc(sizeof(struct camac_single), GFP_KERNEL);
		if(cam_sing == NULL) {
			err(__FUNCTION__ " Out of Memory\n");
			retval = -ENOMEM;
		}
		if( copy_from_user(cam_sing, (void *) arg, 
				   sizeof(struct camac_single)) ){
			retval = -EFAULT;
			err(__FUNCTION__ " copy_from_user failed");
			goto exit;
		}

		retval = do_camac_single( ccusb, cam_sing, word_siz );
		if( copy_to_user((void*)arg, cam_sing, 
				 sizeof(struct camac_single)) ) {
			retval = -EFAULT;
			err(__FUNCTION__ " copy_to_user failed");
			goto exit;		
		}
		break;
	}
	case CAM_BLKW:
	{
		size_t word_size = 2;
		size_t trans_size = 0;
		unsigned int func;

		dbg(__FUNCTION__ " CAM_BLKW");

		cam_blk = kmalloc(sizeof(struct camac_block), GFP_KERNEL);
		if(cam_blk == NULL) {
			err(__FUNCTION__ " Out of Memory\n");
			retval = -ENOMEM;
		}

		if( copy_from_user(cam_blk, (void *) arg, 
				   sizeof(struct camac_block)) ){
			retval = -EFAULT;
			err(__FUNCTION__ " copy_from_user failed");
			goto exit;
		}
		
		func = cam_blk->naf[2];
		trans_size = do_camac_block( ccusb, cam_blk, word_size );
		cam_blk->retlen = (unsigned int)trans_size;

		if( func >= 0 && func < 8 ){
			if( copy_to_user((unsigned short *)cam_blk->data, 
					 ccusb->bulk_in_buffer, 
					 trans_size) ){
				retval = -EFAULT;
				err(__FUNCTION__ " copy_from_user failed");
				goto exit;
			}
		}
		if( copy_to_user((void *)arg, cam_blk, 
				 sizeof(struct camac_block)) ){
			retval = -EFAULT;
			err(__FUNCTION__ " copy_to_user failed");
			goto exit;
		}
		break;
	}
	case CAM_BLKL:
	{
		size_t word_size = 3;
		size_t trans_size = 0;
		unsigned int func;

		dbg(__FUNCTION__ " CAM_BLKL");

		cam_blk = kmalloc(sizeof(struct camac_block), GFP_KERNEL);
		if(cam_blk == NULL) {
			err(__FUNCTION__ " Out of Memory\n");
			retval = -ENOMEM;
		}

		if( copy_from_user(cam_blk, (void *) arg, 
				   sizeof(struct camac_block)) ){
			retval = -EFAULT;
			err(__FUNCTION__ " copy_from_user failed");
			goto exit;
		}
		func = cam_blk->naf[2];

		trans_size = do_camac_block( ccusb, cam_blk, word_size );
		cam_blk->retlen = (unsigned int)trans_size;

		if( func >= 0 && func < 8 ){
			if( copy_to_user((unsigned short *)cam_blk->data, 
					 ccusb->bulk_in_buffer, 
					 trans_size) ){
				retval = -EFAULT;
				err(__FUNCTION__ " copy_to_user failed");
				goto exit;
			}
		}
		if( copy_to_user((void *)arg, cam_blk, 
				 sizeof(struct camac_block)) ){
			retval = -EFAULT;
			err(__FUNCTION__ " copy_to_user failed");
			goto exit;
		}
		break;
	}
	case CAM_INIT:
	{
		int ret;
		unsigned char camac_com = INITIALIZE;

		dbg(__FUNCTION__ " CAM_INIT");
		ret = usb_write_wait(ccusb, &camac_com, sizeof(camac_com), 
				     ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " write to USB failed\n");
			retval = -EFAULT;
			goto exit;
		}

		ret = read_status(ccusb);
		if (ret < 0) {
			err(__FUNCTION__ " stat = %x\n", ret);
		}
		break; 
	}
	case CAM_CLEAR:
	{
		int ret;
		unsigned char camac_com = CLEAR;

		dbg(__FUNCTION__ " CAM_CLEAR");
		ret = usb_write_wait(ccusb, &camac_com, sizeof(camac_com),
				     ccusb->single_out_ep_addr);
		if(!ret){
			retval = -EFAULT;
			err(__FUNCTION__ " write to USB failed\n");
			goto exit;
		}
		ret = read_status(ccusb);
		if (ret < 0) {
			err(__FUNCTION__ " stat = %x\n", ret);
		}
		break;
	}
	case CAM_SETI:
	{
		int ret;
		unsigned char camac_com = SET_INH;

		dbg(__FUNCTION__ " CAM_SETI");
		ret = usb_write_wait(ccusb, &camac_com, sizeof(camac_com), 
				     ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " write to USB failed\n");
			retval = -EFAULT;
			goto exit;
		}
		ret = read_status(ccusb);
		if (ret < 0) {
			err(__FUNCTION__ " stat = %x\n", ret);
		}

		break;
	}
	case CAM_CLRI:
	{
		int ret;
		unsigned char camac_com = CLR_INH;

		dbg(__FUNCTION__ " CAM_CLRI");
		ret = usb_write_wait(ccusb, &camac_com, sizeof(camac_com), 
				     ccusb->single_out_ep_addr);
		if(!ret){
			err(__FUNCTION__ " write to USB failed\n");
			retval = -EFAULT;
			goto exit;
		}
		ret = read_status(ccusb);
		if (ret < 0) {
			err(__FUNCTION__ " stat = %x\n", ret);
		}

		break;
	}
	case CAM_SETLAM:
	{
		int ret;
		unsigned char camac_com = SET_LAM;

		dbg(__FUNCTION__ " CAM_SETLAM");
		ret = usb_write_wait(ccusb, &camac_com, sizeof(camac_com), 
				     ccusb->single_out_ep_addr);
		if(!ret){
			retval = -EFAULT;
			err(__FUNCTION__ " write to USB failed\n");
			goto exit;
		}
		ret = read_status(ccusb);
		if (ret < 0) {
			err(__FUNCTION__ " stat = %x\n", ret);
		}
		break;
	}
	case CAM_CLRLAM:
	{
		int ret;
		unsigned char camac_com = CLR_LAM;

		dbg(__FUNCTION__ " CAM_CLRLAM");
		ret = usb_write_wait(ccusb, &camac_com, sizeof(camac_com), 
				     ccusb->single_out_ep_addr);
		if(!ret){
			retval = -EFAULT;
			err(__FUNCTION__ "write to USB failed\n");
			goto exit;
		}
		ret = read_status(ccusb);
		if (ret < 0) {
			err(__FUNCTION__ " stat = %x\n", ret);
		}
		break;
	}
	case CAM_READLAM:
	{
		int ret;
		unsigned char camac_com = READ_LAM;
		unsigned char camac_lam_stat[3];

		dbg(__FUNCTION__ " CAM_READLAM");
		ret = usb_write(ccusb, &camac_com, sizeof(camac_com), 
				ccusb->single_out_ep_addr);
		if(!ret){
			retval = -EFAULT;
			err(__FUNCTION__ " write to USB failed\n");
			goto exit;
		}
		ret = usb_read(ccusb, camac_lam_stat, sizeof(camac_lam_stat), 
			       ccusb->single_in_ep_addr);
		if(ret <= 0){
			retval = -EFAULT;
			err(__FUNCTION__ " read from USB failed\n");
			goto exit;
		}
		ret = read_status(ccusb);
		if (ret < 0) {
			err(__FUNCTION__ " stat = %x\n", ret);
		}
		retval =  (camac_lam_stat[2] << 16) 
			+ (camac_lam_stat[1] << 8) 
			+  camac_lam_stat[0];
		dbg(__FUNCTION__ " LAM[0] %x\n", camac_lam_stat[0]);
		dbg(__FUNCTION__ " LAM[1] %x\n", camac_lam_stat[1]);
		dbg(__FUNCTION__ " LAM[2] %x\n", camac_lam_stat[2]);

		break;
	}
	case CAM_READID:
	{
		dbg(__FUNCTION__ " CAM_READID");
		retval = read_id(ccusb);
		dbg(__FUNCTION__ " retval = %d", retval);
		break;
	}
	default:
		retval =  -ENOIOCTLCMD;
		goto exit;
	}
	
 exit:
	if(cam_blk != NULL){
		kfree(cam_blk);
		dbg(__FUNCTION__ " cam_blk free");
	}
	if(cam_sing != NULL){
		kfree(cam_sing);
		dbg(__FUNCTION__ " cam_sing free");
	}
	up (&ccusb->sem);

	return retval;
}
/*--------------------------------------------------------------------------*/
static int
ccusb_open(struct inode *inode, struct file *file)
{
	struct ccusb_data *ccusb = NULL;
	int subminor;
	int retval = 0;

	dbg(__FUNCTION__);

	subminor = MINOR(inode->i_rdev) - USB_CCUSB_MINOR_BASE;
        dbg(__FUNCTION__ " subminor = %d\n", subminor);
	if( (subminor < 0) || (subminor >= MAX_DEVICES) ){
		err(__FUNCTION__ " NO Device\n");
		return -ENODEV;
	}
	
	MOD_INC_USE_COUNT;

	down(&minor_table_mutex);

	ccusb = minor_table[subminor];
	if( ccusb == NULL ) {
		up (&minor_table_mutex);
		MOD_DEC_USE_COUNT;
		err(__FUNCTION__ " ccusb not exist\n");
		return -ENODEV;
	}
	
	down (&ccusb->sem);

	up (&minor_table_mutex);

	++ccusb->open_count;
	dbg(__FUNCTION__ " ccusb->open_count = %d\n", ccusb->open_count);
	file->private_data = ccusb;

	up (&ccusb->sem);
	return retval;
}
/*--------------------------------------------------------------------------*/
static int 
ccusb_close(struct inode *inode, struct file *file)
{
	struct ccusb_data *ccusb = NULL;
	int retval = 0;

	dbg(__FUNCTION__);
	ccusb = (struct ccusb_data *) file->private_data;
	if(ccusb == NULL) {
		err(__FUNCTION__ " ccusb not exist\n");
		return -ENODEV;
	}

	down (&minor_table_mutex);
	down (&ccusb->sem);

	if(ccusb->open_count <= 0) {
		err(__FUNCTION__ " Device Not Opened\n");
		retval = -ENODEV;
		goto exit_not_opened;
	}
	if(ccusb->udev == NULL) {
		up (&ccusb->sem);
		ccusb_delete (ccusb);
		MOD_DEC_USE_COUNT;
		up (&minor_table_mutex);
		return 0;
	}

	--ccusb->open_count;
	dbg(__FUNCTION__ " ccusb->open_count = %d\n", ccusb->open_count);
	if(ccusb->open_count <= 0) {
		usb_unlink_urb (ccusb->bulk_write_urb);
		//usb_unlink_urb (ccusb->block_write_urb);
		ccusb->open_count = 0;
	}

	MOD_DEC_USE_COUNT;

 exit_not_opened:
	up (&ccusb->sem);
	up (&minor_table_mutex);

	return retval;
}
/*--------------------------------------------------------------------------*/
static void ccusb_write_bulk_callback (struct urb *urb)
{
	//struct ccusb_data *ccusb = (struct ccusb_data *)urb->context;

	if ((urb->status != -ENOENT) && (urb->status != -ECONNRESET)) {
	  //printk("urb->status = %d\n", urb->status);
		return;
	}
	dbg(__FUNCTION__ " urb->status = %d\n", urb->status);
	return;
}
/*--------------------------------------------------------------------------*/
static struct file_operations
usb_ccusb_fops = {
	owner: THIS_MODULE,
	ioctl: ccusb_ioctl,
	open:  ccusb_open,
	release: ccusb_close,
};
/*--------------------------------------------------------------------------*/
static void *probe_ccusb(struct usb_device *udev, unsigned int ifnum,
                          const struct usb_device_id *id)
{
        struct ccusb_data                 *ccusb = NULL;
        struct usb_interface              *interface;
        struct usb_interface_descriptor   *if_desc;
        struct usb_endpoint_descriptor    *endpoint;
	int                                adapter_id;
        int                                buffer_size;

	dbg(__FUNCTION__);

        /* See if the device offered us matches what we can accept */
        if ((udev->descriptor.idVendor != USB_CCUSB_VENDOR_ID) ||
            (udev->descriptor.idProduct != USB_CCUSB_PRODUCT_ID)) {
                err(__FUNCTION__ " Unknown Vender ID vid=0x%04X(%u) pid=0x%04X(%u)\n",
                       udev->descriptor.idVendor, udev->descriptor.idVendor, 
                       udev->descriptor.idProduct, udev->descriptor.idProduct );
                return NULL;
        }
        dbg(__FUNCTION__ " Vender ID vid=0x%04X(%u) pid=0x%04X(%u)\n", 
	    udev->descriptor.idVendor, udev->descriptor.idVendor, 
	    udev->descriptor.idProduct, udev->descriptor.idProduct );

	/* select a "subminor" number (part of a minor number) */
        down (&minor_table_mutex);
        for (adapter_id = 0; adapter_id < MAX_DEVICES; ++adapter_id) {
                if (minor_table[adapter_id] == NULL)
                        break;
		else
			dbg(__FUNCTION__ " Adapter_id.= %d exists\n", adapter_id);
        }
        
        if (adapter_id >= MAX_DEVICES) {
                err(__FUNCTION__ " Too many devices plugged in, can not handle this one.");
                goto exit;
        }

        ccusb = kmalloc(sizeof(struct ccusb_data), GFP_KERNEL);
        if(ccusb == NULL ) {
                err(__FUNCTION__ " No memory!");
                goto exit;
        }

	interface = &udev->actconfig->interface[ifnum];

	init_MUTEX (&ccusb->sem);
	ccusb->udev = udev;
	ccusb->interface = interface;
	//ccusb->minor = minor;
	ccusb->adapter_id = adapter_id;
	ccusb->open_count = 0;

        /*
         *   Endpoint in alt. setting(0)
	 *   
	 *     Direction                 EP#   Address
	 *   -----------------------------------------
	 *    IN1(CAMAC Single Read)   :  0  :  0x81
	 *    IN2(CAMAC Block  Read)   :  1  :  0x82
	 *   OUT1(CAMAC Single Write)  :  2  :  0x01
	 *   OUT2(CAMAC Block  Write)  :  3  :  0x02
         */

	if_desc = &interface->altsetting[0];
	dbg(__FUNCTION__ " if_desc->bNumEndpoints = %x\n", 
	    if_desc->bNumEndpoints);

	endpoint = &if_desc->endpoint[0];
	buffer_size = endpoint->wMaxPacketSize;
	ccusb->single_in_ep_addr = endpoint->bEndpointAddress;

	endpoint = &if_desc->endpoint[1];
	ccusb->block_in_ep_addr = endpoint->bEndpointAddress;
	ccusb->bulk_in_buffer = kmalloc ( MAX_TRANS_BYTES, GFP_KERNEL);

	if (!ccusb->bulk_in_buffer) {
		err(__FUNCTION__ " Couldn't allocate bulk_in_buffer");
		goto error;
	}

	endpoint = &if_desc->endpoint[2];
	ccusb->bulk_write_urb = usb_alloc_urb(0);
	if (!ccusb->bulk_write_urb) {
		err(__FUNCTION__ " No free urbs available");
		goto error;
	}

	ccusb->single_out_ep_addr = endpoint->bEndpointAddress;

	endpoint = &if_desc->endpoint[3];

	ccusb->block_out_ep_addr = endpoint->bEndpointAddress;
	ccusb->bulk_out_buffer = kmalloc (MAX_TRANS_BYTES, GFP_KERNEL);

	/* Read the adapter's ID */
	adapter_id = read_id(ccusb);
	if(adapter_id < 0) {
		err(__FUNCTION__ " Could not read adapter's ID: %d\n", adapter_id);
		goto error;
	}

	ccusb->adapter_id = adapter_id;
	minor_table[adapter_id] = ccusb;
	info("TOYO CC/USB ID#%d connected.", adapter_id);

	dbg(__FUNCTION__ " CAMAC Single IN  EP Addr. = %x", ccusb->single_in_ep_addr);
	dbg(__FUNCTION__ " CAMAC Single OUT EP Addr. = %x", ccusb->single_out_ep_addr);
	dbg(__FUNCTION__ " CAMAC Block  IN  EP Addr. = %x", ccusb->block_in_ep_addr);
	dbg(__FUNCTION__ " CAMAC Block  OUT EP Addr. = %x", ccusb->block_out_ep_addr);
        goto exit;

error:
        ccusb_delete (ccusb);
        ccusb = NULL;

exit:
        up (&minor_table_mutex);

        return ccusb;

}
/*--------------------------------------------------------------------------*/
static void
disconnect_ccusb(struct usb_device *udev, void *ptr)
{
        struct ccusb_data *ccusb;
	unsigned char adapter_id;

	dbg(__FUNCTION__);
	ccusb = (struct ccusb_data *)ptr;

	down (&minor_table_mutex);
	down (&ccusb->sem);

	adapter_id = ccusb->adapter_id;

	if (!ccusb->open_count) {
		ccusb_delete (ccusb);
	} else {
		ccusb->udev = NULL;
		up (&ccusb->sem);
	}

        info("TOYO CC/USB ID#%d disconnected.", adapter_id);
	up (&minor_table_mutex);
}

/*--------------------------------------------------------------------------*/
static struct usb_driver
driver_table = {
        name:       "ccusb",
        probe:      probe_ccusb,
        disconnect: disconnect_ccusb,
        fops:       &usb_ccusb_fops,
        minor:      USB_CCUSB_MINOR_BASE,
};
int __init
usb_ccusb_module_init(void)
{
  //create_proc_read_entry ("driver/ccusb", 0, 0, ccusb_read_proc, NULL);
	return usb_register(&driver_table) < 0 ? -1 : 0;
}
/*--------------------------------------------------------------------------*/
void __exit
usb_ccusb_module_cleanup(void)
{
	//remove_proc_entry ("driver/ccusb", NULL);
	usb_deregister(&driver_table);
}
/*--------------------------------------------------------------------------*/
MODULE_AUTHOR("Kazuo Nakayoshi ");
MODULE_DESCRIPTION("USB Device Driver for CAMAC CC/7700 USB Adapter");
MODULE_LICENSE("GPL");
/*--------------------------------------------------------------------------*/
module_init (usb_ccusb_module_init);
module_exit (usb_ccusb_module_cleanup);
/*--------------------------------------------------------------------------*/
